module Eval where

import Poly

import Unbound.LocallyNameless

import Control.Monad.Trans.Class

import qualified Data.Map as Map

isValue :: Exp -> Bool
isValue Unit = True
isValue (Const _) = True
isValue (Lam _) = True
isValue (StoreLoc _) = True
isValue _ = False

step :: Exp -> Store -> Maybe (Exp, Store)
step expr = runFreshMT . step' expr
  where step' (Seq Unit e2) store = return (e2, store)
        step' (Seq e1 e2) store = do
          (e1', store') <- step' e1 store
          return (Seq e1' e2, store')
        step' (App lam@(Lam bnd) e2) store
          | isValue e2 = do
              (x,body) <- unbind bnd
              return $ (subst x e2 body, store)
          | otherwise = do
              (e2', store') <- step' e2 store
              return (App lam e2', store')
        step' (App e1 e2) store = do
          (e1', store') <- step' e1 store
          return (App e1' e2, store')
        step' (Let bnd) store = do
          ((x,e), e2) <- unbind bnd
          let e1 = unembed e
          if isValue e1
            then return (subst x e1 e2, store)
            else do (e1', store') <- step' e1 store
                    return (Let (bind (x, embed e1') e2), store')
        step' (Ref e) store
          | isValue e = let loc = length store
                        in return (StoreLoc loc, Map.insert loc e store)
          | otherwise = do
              (e', store') <- step' e store
              return (Ref e', store')
        step' (Deref (StoreLoc l)) store =
          case Map.lookup l store of
            Just e' -> return (e', store)
            Nothing -> lift Nothing
        step' (Deref e) store = do
          (e', store') <- step' e store
          return (Deref e', store')
        step' (Assign (StoreLoc l) e) store
          | isValue e =
              return (Unit, Map.update (const (Just e)) l store)
          | otherwise = do
              (e', store') <- step' e store
              return (Assign (StoreLoc l) e', store')
        step' (Assign e1 e2) store = do
          (e1', store') <- step' e1 store
          return (Assign e1' e2, store')
        step' (Succ (Const i)) store =
          return (Const (i+1), store)
        step' (Succ e) store = do
          (e', store') <- step' e store
          return (Succ e', store')
        step' _ _ = lift Nothing

eval :: Exp -> Store -> (Exp, Store)
eval e store =
  case step e store of
    Just (e', store') -> eval e' store'
    _ -> (e,store)
