{-# LANGUAGE MultiParamTypeClasses
           , TemplateHaskell
           , ScopedTypeVariables
           , FlexibleInstances
           , FlexibleContexts
           , UndecidableInstances
  #-}

module Poly where

import Unbound.LocallyNameless

import Data.Map (Map)

type TyName = Name Ty
type ExpName = Name Exp

type Location = Int
type Store = Map Location Exp

data Ty = TyInt
        | TyUnit
        | TyVar TyName
        | TyFun Ty Ty
        | TyRef Ty

newtype TyScheme = TyScheme (Bind [TyName] Ty)

data Exp = Var ExpName
         | App Exp Exp
         | Lam (Bind ExpName Exp)
         | Seq Exp Exp
         | Let (Bind (ExpName, Embed Exp) Exp)
         | Ref Exp
         | Deref Exp
         | Assign Exp Exp
         | StoreLoc Int
         | Const Int
         | Succ Exp
         | Unit
         deriving Show

-- naive pretty printing of types

instance Show Ty where
  show TyUnit = "unit"
  show TyInt = "int"
  show (TyVar x) = show x
  show (TyRef ty) = "ref " ++ show ty
  show (TyFun ty1 ty2) = show ty1 ++ " -> " ++ show ty2

instance Show TyScheme where
  show (TyScheme tsc) = runFreshM $ do
    (vars, ty) <- unbind tsc
    return $ forallPrefix vars ++ show ty

forallPrefix :: [TyName] -> String
forallPrefix [] = ""
forallPrefix xs = "forall " ++ show xs ++ ". "

-- unbound stuff

$(derive [''Ty, ''TyScheme, ''Exp])

instance Alpha Ty
instance Alpha TyScheme
instance Alpha Exp

instance Subst Ty Ty where
  isvar (TyVar x) = Just (SubstName x)
  isvar _         = Nothing
instance Subst Ty TyScheme  where
instance Subst Exp Ty where
instance Subst Exp Exp where
  isvar (Var x) = Just (SubstName x)
  isvar _       = Nothing

-- equivalence is alpha-equivalence

instance Eq Ty where
  (==) = aeq
instance Eq TyScheme where
  (==) = aeq
instance Eq Exp where
  (==) = aeq
