module Properties where

import Poly
import Eval
import Typing

import Unbound.LocallyNameless hiding (apply)
import qualified Data.Map as Map
import Data.Maybe (isJust)

canStep :: Exp -> Store -> Bool
canStep e = isJust . step e

progressCheck :: Exp -> Bool
progressCheck e1
  | wellTyped e1 = checkStepByStep e1 Map.empty
  | otherwise = True
  where checkStepByStep e s =
          case step e s of
            Nothing -> isValue e
            Just (e',s') -> checkStepByStep e' s'

{- NOTE
   we are discrading counterexamples here: if the term obtained after eval is ill-typed,
   we simply return true. The problem is that if we were to consider these terms as
   counterexamples, then our type system would become more complex because we would have
   to carry around a store typing. The current implementation is similar to the one
   provided with TAPL, in which it is assumed that we are type-checking programs written
   by the user in which store locations cannot appear. Nonetheless, our value restriction
   counterexample is not in this class of discarded counterexamples, hence we are still
   searching for something that exists and can be found.
-}
preservationCheck :: Exp -> Bool
preservationCheck e
  | canStep e Map.empty = maybe True id $ do
      tys <- infer [] e
      let (e', _) = eval e Map.empty
      tys' <- infer [] e'
      return (tys `unifiesWith` tys')
  | otherwise = True

unifiesWith :: TyScheme -> TyScheme -> Bool
unifiesWith (TyScheme ts1) (TyScheme ts2) = isJust $ runFreshM $ do
  (_, ty1) <- unbind ts1
  (_, ty2) <- unbind ts2
  return (mgu ty1 ty2)
