module SanityCheckSpec (spec) where

import Test.Hspec

import Poly
import Eval
import Typing 
import Properties

import Data.Map as Map

import Unbound.LocallyNameless

spec = do
  describe "Value Restriction (Preservation)" valueRestrictionPreservation
  describe "Value Restriction (Progress)" valueRestrictionProgress

valueRestrictionPreservation = do
  let ctx1 = Let (bind
                   (s2n "x",
                    embed $ Ref $ Lam (bind (s2n "x") (Var (s2n "x"))))
                   (Seq
                     (Assign (Var (s2n "x")) (Lam (bind (s2n "x") Unit)))
                     (App (Deref (Var (s2n "x"))) (Const 0))))

  it "the counterexample #1 type-checks" $ do
    ctx1 `shouldSatisfy` wellTyped

  it "the counterexample #1 breaks preservation (with eval instead of step)" $ do
    ctx1 `shouldNotSatisfy` preservationCheck

valueRestrictionProgress = do
  let ctx1 = Let (bind
                   (s2n "x",
                    embed $ Ref $ Lam (bind (s2n "x") (Var (s2n "x"))))
                   (Seq
                     (Assign (Var (s2n "x")) (Lam (bind (s2n "x") (Succ (Var (s2n "x"))))))
                     (App (Deref (Var (s2n "x"))) Unit)))

  let ctx2 = Let (bind
                   (s2n "r",
                    embed $ Ref $ Lam (bind (s2n "x") (Var (s2n "x"))))
                   (Let (bind
                        (s2n "u",
                         embed $ Assign (Var (s2n "r")) (Lam (bind (s2n "y") (Ref (Deref (Var (s2n "y")))))))
                        (App (Deref (Var (s2n "r"))) Unit))))

  it "the counterexample #1 type-checks" $ do
    ctx1 `shouldSatisfy` wellTyped

  it "the counterexample #1 breaks progress (with eval instead of step)" $ do
    ctx1 `shouldNotSatisfy` progressCheck

  it "the counterexample #2 type-checks" $ do
    ctx2 `shouldSatisfy` wellTyped

  it "the counterexample #2 breaks progress (with eval instead of step)" $ do
    ctx2 `shouldNotSatisfy` progressCheck

